#!/usr/bin/env python
'''
Custom operations for EMBL submission preparation tool
'''

#####################
# IMPORT OPERATIONS #
#####################

import GlobalVariables as GlobVars

from operator import itemgetter
from itertools import groupby
from Bio import SeqFeature
from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import ExactPosition, FeatureLocation, CompoundLocation

###############
# AUTHOR INFO #
###############

__author__ = 'Michael Gruenstaeudl <m.gruenstaeudl@fu-berlin.de>'
__copyright__ = 'Copyright (C) 2016-2019 Michael Gruenstaeudl'
__info__ = 'annonex2embl'
__version__ = '2019.10.16.1700'

#############
# DEBUGGING #
#############

#import ipdb
#ipdb.set_trace()

###########
# CLASSES #
###########

class GenerateFeatLoc:
    ''' This class contains functions to generate or manipulate
        SeqFeature location objects.
    '''

    def __init__(self):
        pass

    @staticmethod
    def _exact(csrange):
        ''' An internal static function to generate an exact feature
            location. '''
        start_pos = csrange[0]
        stop_pos = csrange[-1] + 1
        start_exact = ExactPosition(start_pos)
        stop_exact = ExactPosition(stop_pos)
        return FeatureLocation(start_exact, stop_exact)

    @staticmethod
    def _extract_contiguous_subsets(compound_integer_range):
        ''' An internal static function to extract all contiguous
            integer ranges from compound integer range.
        '''
#        Examples:
#            Example 1:
#            >>> compound_integer_range = [1,2,3,7,8,9]
#            >>> _extract_contiguous_subsets(compound_integer_range)
#            Out: [[1, 2, 3], [7, 8, 9]]

        outlist = []
        for k, g in groupby(enumerate(compound_integer_range),
                            lambda i_x: i_x[0] - i_x[1]):
            outlist.append(list(map(itemgetter(1), g)))
        return outlist

    def make_location(self, charset_range):
        ''' This function goes through a decision tree and generates
            fitting feature locations.
        Args:
            charset_range (list): a list of index positions, example: [1,2,3,8,9 ...]
        Returns:
            FeatureLocation (obj):  A SeqFeature location object; either a
                                    FeatureLocation or a CompoundLocation
        Raises:
            -
        '''
        contiguous_ranges = GenerateFeatLoc._extract_contiguous_subsets(
            charset_range)
        # Convert each contiguous range into an exact feature location
        for countr, rnge in enumerate(contiguous_ranges):
            contiguous_ranges[countr] = GenerateFeatLoc._exact(rnge)
        if len(contiguous_ranges) > 1:
            return CompoundLocation(contiguous_ranges)
        else:
            return contiguous_ranges[0]

    def make_location_complement(self, location_object):
        location_object._set_strand(-1)
        return location_object

    def make_start_fuzzy(self, location_object):
        ''' This function makes the start position of location
            objects fuzzy.
        '''
        orient = location_object._get_strand()
        if hasattr(location_object, 'parts'):
            if len(location_object.parts) == 1:
                if orient == -1:  # If feature reverse, fuzzy start must change with it
                    new_start_pos = SeqFeature.BeforePosition(
                        location_object.end)
                    location_object = SeqFeature.FeatureLocation(
                        location_object.start, new_start_pos)
                else:
                    new_start_pos = SeqFeature.BeforePosition(
                        location_object.start)
                    location_object = SeqFeature.FeatureLocation(
                        new_start_pos, location_object.end)
            if len(location_object.parts) > 1:
                if orient == -1:
                    new_start_pos = SeqFeature.BeforePosition(
                        location_object.parts[0].end)
                    location_object.parts[0] = SeqFeature.FeatureLocation(
                        location_object.parts[0].start, new_start_pos)
                else:
                    new_start_pos = SeqFeature.BeforePosition(
                        location_object.parts[0].start)
                    location_object.parts[0] = SeqFeature.FeatureLocation(
                        new_start_pos, location_object.parts[0].end)
        location_object._set_strand(orient)
        return location_object

    def make_end_fuzzy(self, location_object):
        ''' This function makes the end position of location
            objects fuzzy.
        '''

#        Examples:
#            Example 1:
#                >>> from Bio import SeqFeature
#                >>> start_pos = SeqFeature.ExactPosition(5)
#                >>> end_pos = SeqFeature.ExactPosition(9)
#                >>> location_object = SeqFeature.FeatureLocation(start_pos, end_pos)
#                >>> location_object
#                Out: FeatureLocation(ExactPosition(5), ExactPosition(9))
#                >>> new_loc = GenerateFeatLoc().make_end_fuzzy(location_object)
#                >>> new_loc
#                Out: FeatureLocation(ExactPosition(5), AfterPosition(9))
#
#            Example 2:
#                >>> from Bio import SeqFeature
#                >>> csrange = [1,2,3,7,8]
#                >>> location_object = GenerateFeatLoc().make_location(csrange)
#                >>> location_object
#                Out: CompoundLocation([FeatureLocation(ExactPosition(1),
#                ExactPosition(4)), FeatureLocation(ExactPosition(7),
#                ExactPosition(9))], 'join')
#                >>> new_loc = GenerateFeatLoc().make_end_fuzzy(location_object)
#                >>> new_loc
# Out: CompoundLocation([FeatureLocation(ExactPosition(1),
# ExactPosition(4)), FeatureLocation(ExactPosition(7), AfterPosition(9))],
# 'join')
        orient = location_object._get_strand()
        if hasattr(location_object, 'parts'):
            if len(location_object.parts) == 1:
                if orient == -1: # If feature reverse, fuzzy end must change with it
                    new_end_pos = SeqFeature.BeforePosition(location_object.start)
                    location_object = SeqFeature.FeatureLocation(
                        new_end_pos, location_object.end)
                else:
                    new_end_pos = SeqFeature.AfterPosition(location_object.end)
                    location_object = SeqFeature.FeatureLocation(
                        location_object.start, new_end_pos)
            if len(location_object.parts) > 1:
                if orient == -1:
                    new_end_pos = SeqFeature.BeforePosition(
                        location_object.parts[-1].start)
                    location_object.parts[-1] = SeqFeature.FeatureLocation(
                        new_end_pos, location_object.parts[-1].end)
                else:
                    new_end_pos = SeqFeature.AfterPosition(
                        location_object.parts[-1].end)
                    location_object.parts[-1] = SeqFeature.FeatureLocation(
                        location_object.parts[-1].start, new_end_pos)
        location_object._set_strand(orient)

        return location_object


class GenerateSeqFeature:
    ''' This class contains functions to generate SeqFeatures. '''

    def __init__(self):
        pass

    def source_feat(self, full_len, quals, charset_names):
        ''' This function generates the SeqFeature source for a
            SeqRecord. The SeqFeature source is critical for
            submissions to EMBL or GenBank, as it contains all the
            relevant info on collection locality, herbarium voucher,
            etc. It also provides info on which translation table is
            used if a CDS sequence feature is encountered among the
            gene names.
        Args:
            full_len (int): the full length of the seq in question;
                            example: 509
            quals (dict):   a dictionary of qualifiers; example:
                            {'isolate': 'taxon_B', 'country': 'Ecuador'}
            charset_names (list): a list of gene names; example:
                            ['foo_gene', 'foo_CDS']
        Returns:
            SeqFeature (obj):   A SeqFeature object
        Raises:
            [currently nothing]
        '''
        full_index = list(range(0, full_len))
        feature_loc = GenerateFeatLoc().make_location(full_index)
        quals['mol_type'] = "genomic DNA"
        source_feature = SeqFeature.SeqFeature(
            feature_loc,
            id='source',
            type='source',
            qualifiers=quals)
        return source_feature

    def regular_feat(self, feature_name, feature_type, feature_orient, feature_loc,
                     transl_table, feature_seq, feature_product=None):
        ''' This function generates a regular SeqFeature for a SeqRecord.
        Args:
            feature_name (str):  usually a gene symbol; example: 'matK'
            feature_type (str):  an identifier as to the type of feature;
                                 example: 'intron'
            feature_orient (str): a string defining
            feature_loc (object): a SeqFeature object specifying a simple
                                  or compund location on a DNA string
            transl_table (int): an integer; example: 11 (for bacterial code)
            feature_product (str): the product of the feature in question;
                                   example: 'maturase K'
            feature_seq (str): nucleotide sequence sequence
        Returns:
            SeqFeature (obj):   A SeqFeature object
        Raises:
            -
        '''
        # 1. Define the annotation type
        if feature_type not in GlobVars.nex2ena_valid_INSDC_featurekeys:
            msg = '%s nex2embl ERROR: Internal error: '\
                  'Name of feature key not passed correctly.'
            warnings.warn(msg)
            raise Exception
        # 2. Generate qualifiers
        quals = {}
        if feature_name:
            quals = {'note': feature_name}
        # 3. If a coding feature, add special qualifiers
        if feature_product:
            if feature_type == 'CDS' or feature_type == 'gene':
                quals['product'] = feature_product
        if feature_type == 'CDS':
            quals['transl_table'] = transl_table
            # Add a function to add "/codon_start=1" in CDS feature,
            # if start and stop position of feature is uncertain
            # (i.e., <100..>200).
            if (not feature_seq.startswith(GlobVars.nex2ena_start_codon)) or (all([not feature_seq.endswith(c) for c in GlobVars.nex2ena_stop_codons])):
                quals['codon_start'] = 1
        if feature_type == 'gap':
            quals['estimated_length'] = str(feature_loc.end.position-feature_loc.start.position)
            #quals['estimated_length'] = str(feature_loc.end.real-feature_loc.start.real+1)
        # 4. Add a function to read in if a charset is forward or reverse and to adjust the info in the feature table.
        if feature_orient == "forw":
            feature_orient = 1
        else:
            feature_orient = -1
        seq_feature = SeqFeature.SeqFeature(
            feature_loc,
            id=feature_name,
            type=feature_type,
            strand=feature_orient,
            qualifiers=quals)
        return seq_feature




class GenerateSeqRecord:
    ''' This class contains functions to generate SeqRecords. '''

    def __init__(self):
        pass

    def base_record(self, current_seq, current_qual, uniq_seqid_col,
                    seq_version, descr_DEline, topology, tax_division,
                    organelle):
        ''' This function generates a base SeqRecord (i.e., the foundation to
            subsequent SeqRecords).
        Args:
            current_seq (str): the DNA sequence of; example: 1
            current_qual (xxx): foobar; example: foobar
            uniq_seqid_col (str): the column label of the .csv-file that
                                  contains info on the sequence names;
                                  example: "isolate"
            seq_version (str):    an integer in string format
            descr_DEline (str):   a text string to be included in the
                                  DE line
            topology (str):       one of the valid ENA topology
                                  specifications
            tax_division (str):   one of the valid ENA taxonomic
                                  divisions
            organelle (str):      one of the valid INDSC organelle
                                  descriptors
        Returns:
            SeqRecord (obj):      A SeqRecord object
        '''
        # 1. Selecting correct sequence line
        uniq_seqid = current_qual[uniq_seqid_col]
        # 2. Generating parse-able ID line
        ID_line = uniq_seqid + '.' + seq_version
        # Note to line above: seq_version is parsed internally from
        # ID-line when EMBL format is written
        try:
            org_name = current_qual['organism']
        except Exception:
            org_name = 'undetermined organism'
        # 3. Generating DE line
        descr_DEline = descr_DEline.replace('"', '')
        DE_line = ' '.join([org_name, descr_DEline + ',', 'isolate',
                            uniq_seqid])
        # 4. Set up new seq record
        seq_record = SeqRecord(current_seq, id=ID_line, name=org_name,
                               description=DE_line)
        # 5. Specify the topology of the sequence
        if topology in GlobVars.nex2ena_valid_topologies:
            seq_record.annotations['topology'] = topology
        #else:
        #    seq_record.annotations['topology'] = 'linear'
        # 6. Add ID line info on 'taxonomic division'
        if tax_division in GlobVars.nex2ena_valid_tax_divisions:
            seq_record.annotations['data_file_division'] = tax_division
        #else:
        #    seq_record.annotations['data_file_division'] = 'UNC'
        # 7. Specify the organelle of the sequence
        if organelle in GlobVars.nex2ena_valid_INSDC_organelle:
            seq_record.annotations['organelle'] = organelle
        return seq_record



#!/usr/bin/env python2.7
'''
Command-line execution of annonex2embl
'''

#####################
# IMPORT OPERATIONS #
#####################

import sys
import os

# Add specific directory to sys.path in order to import its modules
# NOTE: THIS RELATIVE IMPORTING IS AMATEURISH.
# NOTE: COULD THE FOLLOWING IMPORT BE REPLACED WITH 'import annonex2embl'?
sys.path.append(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'annonex2embl'))

# IMPORTANT: TFL must be after "sys.path.append"
import Annonex2emblMain
import argparse

###############
# AUTHOR INFO #
###############

__author__ = 'Michael Gruenstaeudl <m.gruenstaeudl@fu-berlin.de>'
__copyright__ = 'Copyright (C) 2016-2019 Michael Gruenstaeudl'
__info__ = 'annonex2embl'
__version__ = '2019.10.16.1700'

#############
# DEBUGGING #
#############

#import ipdb
#ipdb.set_trace()

############
# ARGPARSE #
############

class CLI():

    def __init__(self):
        self.client()

    def client(self):

        parser = argparse.ArgumentParser(description="  --  ".join([__author__, __copyright__, __info__, __version__]))

        ### REQUIRED ###
        parser.add_argument('-n',
                            '--nexus',
                            help='absolute path to infile; infile in NEXUS format; Example: /path_to_input/test.nex',
                            default='/home/username/Desktop/test.nex',
                            required=True)

        parser.add_argument('-c',
                            '--csv',
                            help='absolute path to infile; infile in CSV format; Example: /path_to_input/test.csv',
                            default='/home/username/Desktop/test.csv',
                            required=True)

        parser.add_argument('-d',
                            '--descr',
                            help='text string characterizing the multiple sequence alignment contained in the NEXUS file; Example: "chloroplast trnR-atpA intergenic spacer"',
                            default='a_description_here',
                            required=True)

        parser.add_argument('-e',
                            '--email',
                            help='email address of submitter; Example: "your_email_here@yourmailserver.com"',
                            default='your_email_here@yourmailserver.com',
                            required=True)

        parser.add_argument('-a',
                            '--authors',
                            help='Author names; Example: "LastName1 A.; LastName2 B."',
                            default='LastName1 A.; LastName2 B.',
                            required=True)

        parser.add_argument('-o',
                            '--outfile',
                            help='absolute path to outfile; outfile in EMBL format; Example: /path_to_output/test.embl',
                            default='/home/username/Desktop/test.embl',
                            required=True)

        ### OPTIONAL ###
        parser.add_argument('--manifeststudy',
                            help='the study accession number for the manifest file that the submission shall be linked to; Example: PRJEB00000',
                            default='',
                            required=False)

        parser.add_argument('--manifestdescr',
                            help='a unique, non-interrupted text string for the manifest file characterizing the input alignment; Example: a_unique_description_here',
                            default='',
                            required=False)

        parser.add_argument('--productlookup',
                            help='a logical; shall gene products be looked up via NCBI and their names added to the output flatfile?',
                            default='False',
                            required=False)

        parser.add_argument('--taxonomycheck',
                            help='a logical; shall scientific taxon names be confirmed via the ENA taxonomy service?',
                            default='False',
                            required=False)

        parser.add_argument('--linemask',
                            help='a logical; shall the ID and the AC lines of the output flatfile be masked?',
                            default='False',
                            required=False)

        parser.add_argument('--seqtopol',
                            help='topology of the input sequences; available options: circular or linear',
                            default='linear',
                            required=False)

        parser.add_argument('--taxdivision',
                            help='any of the three letter codes indicating the taxonomic division of the sequences; '\
                            'for details, see: section 3.2 of the EMBL user manual',
                            default='PLN',
                            required=False)

        parser.add_argument('--collabel',
                            #metavar='column specifying sequence names',
                            help='name of the metadata (i.e., CSV file) column that specifies the sequence names',
                            default='isolate',
                            required=False)

        parser.add_argument('--transltable',
                            #metavar='translation table',
                            help='ID number of the translation table used for translating coding regions; '\
                            'for details, see: http://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi',
                            default='11',
                            required=False)

        parser.add_argument('--organelle',
                            #metavar='translation table',
                            help='type of membrane-bound intracellular structure, if any, from which the sequence was obtained; '\
                            'for details, see: http://www.insdc.org/files/feature_table.html',
                            default='plastid',
                            required=False)

        parser.add_argument('--seqvers',
                            #metavar='sequence version',
                            help='an integer',
                            default='1',
                            required=False)

        parser.add_argument('--compress',
                            #metavar='sequence version',
                            help='a logical; shall the output flatfile be compressed (.gz) upon generation?',
                            default='False',
                            required=False)

        parser.add_argument('--version',
                            help='print version number and exit',
                            action='version',
                            version='%(prog)s ' + __version__)

        args = parser.parse_args()

        Annonex2emblMain.annonex2embl(   args.nexus,
                                         args.csv,
                                         args.descr,
                                         args.email,
                                         args.authors,
                                         args.outfile,

                                         args.manifeststudy,
                                         args.manifestdescr,
                                         args.productlookup,
                                         args.taxonomycheck,
                                         args.linemask,
                                         args.seqtopol,
                                         args.taxdivision,
                                         args.collabel,
                                         args.transltable,
                                         args.organelle,
                                         args.seqvers,
                                         args.compress )

########
# MAIN #
########

def start_annonex2embl():
    CLI()



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-

if __name__ == '__main__':
    
    import sys, os
    sys.path.append(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'annonex2embl'))
    from annonex2embl import CLIOps

    CLIOps.start_annonex2embl()



#!/usr/bin/env python
'''
Classes to degap sequences but maintain annotations
'''

#####################
# IMPORT OPERATIONS #
#####################

import warnings

from copy import copy
from itertools import count, groupby

###############
# AUTHOR INFO #
###############

__author__ = 'Michael Gruenstaeudl <m.gruenstaeudl@fu-berlin.de>'
__copyright__ = 'Copyright (C) 2016-2019 Michael Gruenstaeudl'
__info__ = 'annonex2embl'
__version__ = '2019.10.16.1700'

#############
# DEBUGGING #
#############

#import ipdb
#ipdb.set_trace()

# To format warnings in a pretty, readable way:
def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return '\n annonex2embl %s\n' % (message)
warnings.formatwarning = warning_on_one_line

###########
# CLASSES #
###########

class AddGapFeature:
    ''' This class contains a function that identifies a stretch of Ns
        in an input sequence and automatically adds a gap charset in
        its position.
    Args:
        seq (str):      a string that represents an aligned, degapped
                        DNA sequence; example: "ATGNNNC"
        charsets (dict):a dictionary with gene names (str) as keys and lists
                        of nucleotide positions (list) as values; example:
                        {"gene_1":[0,1],"gene_2":[2,3,4]}
    Returns:
        tupl.   The return consists of the input sequence and the
                corresponding charsets (plus a gap charset, if
                appropriate); example: (degapped_seq, degapped_charsets)
    Raises:
        currently nothing
    '''

    def __init__(self, seq, charsets):
        self.seq = seq
        self.charsets = charsets

    def add(self):
        ''' This function was developed while reviewing the following answer on SO:
            https://stackoverflow.com/questions/25211905/determine-length-of-polypurine-tract '''
        seq = self.seq
        charsets = self.charsets
        annotations = copy(charsets)
        gap_indices = [i for i, nucl in enumerate(seq) if nucl=="N"]  # Indexing all 'N' in seq
        gap_ranges = [list(g) for _,g in groupby(gap_indices, key=lambda n, c=count(): n-next(c))]
        if gap_ranges:
            for countr, rnge in enumerate(gap_ranges):
                try:
                    annotations["gap"+str(countr)] = rnge
                except Exception as e:
                    warnings.warn("Cannot process Ns in positions %s." % (','.join(rnge)))
        return seq, annotations

class DegapButMaintainAnno:
    ''' This class contains a function to degap DNA sequences while
        maintaining annotations. Specifically, the functions remove
        dashes from strings while maintaining annotations on these
        strings. Only some of the implementations work if the charsets
        are overlapping.
    Args:
        seq (str):      a string that represents an aligned DNA sequence;
                        example: "ATG-C"
        charsets (dict):a dictionary with gene names (str) as keys and lists
                        of nucleotide positions (list) as values; example:
                        {"gene_1":[0,1],"gene_2":[2,3,4]}
    Returns:
        tupl.   The return consists of the degapped sequence and the
                corresponding degapped charsets; example:
                (degapped_seq, degapped_charsets)
    Raises:
        currently nothing
    '''

    def __init__(self, seq, rmchar, charsets):
        self.seq = seq
        self.rmchar = rmchar
        self.charsets = charsets

    def degap(self):
        ''' This function works on overlapping charsets and is preferable over
        "degap_legacy".
        Source: http://stackoverflow.com/questions/35233714/
        maintaining-overlapping-annotations-while-removing-dashes-from-string
        '''

        seq = self.seq
        rmchar = self.rmchar
        charsets = self.charsets
        annotations = copy(charsets)
        index = seq.find(rmchar)
        while index > -1:  # if any occurrence is found
            for gene_name, indices in list(annotations.items()):
                if index in indices:
                    indices.remove(index)
                annotations[gene_name] = [e-1 if e > index else e
                                          for e in indices]
            seq = seq[:index] + seq[index+1:]
            index = seq.find(rmchar)
        return seq, annotations


class RmAmbigsButMaintainAnno:
    ''' This class removes ambiguous nucleotides from a DNA sequence
        while maintaining the annotations.
    Args:
        seq (str):      a string that represents an aligned DNA sequence;
                        example: "NNATGCNNN"
        charsets (dict):a dictionary with gene names (str) as keys and lists
                        of nucleotide positions (list) as values; example:
                        {"gene_1":[0,1,2,3],"gene_2":[4,5,6,7,8]}
    Returns:
        tupl.   The return consists of the shortened DNA sequence and
                the corresponding shortened charsets; example:
                (shortened_seq, shortened_charsets)
    Raises:
        currently nothing
    '''

    def __init__(self):
        pass

    @staticmethod
    def rm_leadambig(seq, rmchar, charsets):
        ''' This class removes leading ambiguous nucleotides from a DNA
            sequence while maintaining the annotations. '''
        if seq[0] == rmchar:
            lead_stripoff = len(seq)-len(seq.lstrip(rmchar))
            for gene_name, indices in list(charsets.items()):
                indices_shifted = [i-lead_stripoff for i in indices]
                charsets[gene_name] = [i for i in indices_shifted if i >= 0]
            seq = seq[lead_stripoff:]
        return seq, charsets

    @staticmethod
    def rm_trailambig(seq, rmchar, charsets):
        ''' This class removes trailing ambiguous nucleotides from a DNA
            sequence while maintaining the annotations. '''
        if seq[-1] == rmchar:
            trail_stripoff = len(seq.rstrip(rmchar))
            range_stripoff = list(range(trail_stripoff, len(seq)))
            for gene_name, indices in list(charsets.items()):
                indices_new = copy(indices)
                for index in range_stripoff:
                    if index in indices_new:
                        indices_new.remove(index)
                #    else:
                #        warnings.warn("Index %s out of range." %(index))
                charsets[gene_name] = indices_new
            seq = seq[:trail_stripoff]
        return seq, charsets



#!/usr/bin/env python
'''
Custom operations input and output processes
'''

#####################
# IMPORT OPERATIONS #
#####################

import os
import datetime

from csv import DictReader
from Bio.Nexus import Nexus
from Bio import SeqIO

try:
    from StringIO import StringIO ## for Python 2
except ImportError:
    from io import StringIO ## for Python 3

###############
# AUTHOR INFO #
###############

__author__ = 'Michael Gruenstaeudl <m.gruenstaeudl@fu-berlin.de>'
__copyright__ = 'Copyright (C) 2016-2019 Michael Gruenstaeudl'
__info__ = 'annonex2embl'
__version__ = '2019.10.16.1700'

#############
# DEBUGGING #
#############

#import ipdb
#ipdb.set_trace()

###########
# CLASSES #
###########

class Inp:
    ''' This class contains functions to conduct miscellaneous input
        operations.
    Args:
        [specific to function]
    Returns:
        [specific to function]
    Raises:
        -
    '''

    def __init__(self):
        pass

    def extract_fn(self, in_path):
        ''' This function splits a the path from path+filename. '''
        path, fn = os.path.split(in_path)
        return fn

    def repl_fileend(self, fn, new_end):
        ''' This function replaces the file ending (e.g. ".csv") with a
        different ending. '''
        return fn[:fn.rfind('.')] + '.' + new_end


    def parse_csv_file(self, path_to_csv):
        ''' This function parses a csv file. '''
        try:
            reader = DictReader(open(path_to_csv, 'r'), delimiter=',',
                                quotechar='"', skipinitialspace=True)
            a_matrix = list(reader)
        except Exception as e:
            msg = 'ERROR: %s:\n %s' % ('Parsing of '
            '.csv-file unsuccessful', e)
            warnings.warn(msg)
            raise Exception
        return a_matrix

    def parse_nexus_file(self, path_to_nex):
        ''' This function parses a NEXUS file. '''
        try:
            aln = Nexus.Nexus()
            aln.read(path_to_nex)
            charsets = aln.charsets
            matrix = aln.matrix
        except Nexus.NexusError as e:
            print(e)
            raise
        except Exception as e:
            msg = 'ERROR: %s:\n %s' % ('Parsing of '
            '.nex-file unsuccessful', e)
            warnings.warn(msg)
            raise Exception
        return (charsets, matrix)


class Outp:
    ''' This class contains two functions for various output operations.
    Args:
        [specific to function]
    Returns:
        [specific to function]
    Raises:
        -
    '''

    def __init__(self):
        pass

    def write_SeqRecord(self, seq_name, seq_record, author_names, outp_handle, ENAstrict_bool):
        ''' This function writes a seqRecord in EMBL format.
            Upon request (ENAstrict_bool), it masks the ID and AC
            lines as requested by ENA for submissions.
        Args:
            seq_name (str)
            seq_record (obj)
            author_names (str)
            outp_handle (obj)
            ENAstrict_bool(bool)
        Returns:
            currently nothing
        Raises:
            -
        '''

        date_today = datetime.date.today().strftime("%d-%b-%Y").upper()
        SecRecord_handle = StringIO()
        try:
            SeqIO.write(seq_record, SecRecord_handle, 'embl')
        except Exception as e:
            msg = 'ERROR: %s: %s. Did not write to '
            'internal handle.\n %s' % ('Problem with sequence ', seq_name, e)
            warnings.warn(msg)
            raise Exception

        if ENAstrict_bool:
            SecRecord_handle_lines = SecRecord_handle.getvalue().splitlines()
            if SecRecord_handle_lines[0].split()[0] == 'ID':
                ID_line = SecRecord_handle_lines[0]
                ID_line_parts = ID_line.split('; ')
                if len(ID_line_parts) == 7:
                    ID_line_parts = ['XXX' if ID_line_parts.index(p) in
                                     [0, 1, 3, 4, 5, 6] else p for p in ID_line_parts]
                SecRecord_handle_lines[0] = 'ID   ' + '; '.join(ID_line_parts)
            if SecRecord_handle_lines[2].split()[0] == 'AC':
                SecRecord_handle_lines[2] = 'AC   XXX;'
            SecRecord_handle_new = '\n' + '\n'.join(SecRecord_handle_lines)
            SecRecord_handle.truncate(0)
            SecRecord_handle.write(SecRecord_handle_new)
        else:
            pass

        for line in SecRecord_handle.getvalue().split("\n"):
            if line[0:8] == 'FH   Key':
                outp_handle.write("RN   [1]" +
                          "\nRA   " + author_names.replace('"','') +
                          "\nRT   ;" +
                          "\nRL   Submitted (" + date_today + ") to the INSDC." +
                          "\nXX" +
                          "\n")
            outp_handle.write(line)
            if line[0:2] != '//':
                outp_handle.write("\n")
        SecRecord_handle.close()


    def create_manifest(self, path_to_outfile, manifest_study, manifest_name, manifest_flatfile):
        ''' This function writes a manifest file. '''

        manifest_fn = ''.join(path_to_outfile.split('.')[:-1]) + '.manifest'
        with open(manifest_fn, "w") as manifest:
            manifest.write(("STUDY\t %s\n") % (manifest_study))
            manifest.write(("NAME\t %s\n") % (manifest_name))
            manifest.write(("FLATFILE\t %s\n") % (manifest_flatfile))



#!/usr/bin/env python
'''
Classes to parse charset names
'''

#####################
# IMPORT OPERATIONS #
#####################

import GlobalVariables as GlobVars
import sys, os
import pdb
import requests
import unidecode
import warnings

from Bio import Entrez
from collections import Counter

try:
    from urllib.request import urlopen
except Exception as e:
    from urllib2 import urlopen

###############
# AUTHOR INFO #
###############

__author__ = 'Michael Gruenstaeudl <m.gruenstaeudl@fu-berlin.de>'
__copyright__ = 'Copyright (C) 2016-2019 Michael Gruenstaeudl'
__info__ = 'annonex2embl'
__version__ = '2019.10.16.1700'

#############
# DEBUGGING #
#############

#import ipdb
#ipdb.set_trace()

# To format warnings in a pretty, readable way:
def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return '\n annonex2embl %s\n' % (message)
warnings.formatwarning = warning_on_one_line

###########
# CLASSES #
###########

class GetEntrezInfo:
    ''' This class contains functions to obtain gene information from gene
    symbols. '''

    def __init__(self, email_addr):
        self.email_addr = email_addr

    @staticmethod
    def _id_lookup(gene_sym, retmax=10):
        ''' An internal static function to convert a gene symbol to an Entrez ID
            via ESearch.
        Args:
            gene_sym (str): a gene symbol; example: 'psbI'
        Returns:
            entrez_id_list (list): a list of Entrez IDs; example: ['26835430',
                            '26833718', '26833393', ...]
        Raises:
            none
        '''

#        Examples:
#            Example 1: # Default behaviour
#                >>> gene_sym = 'psbI'
#                >>> _id_lookup(gene_sym)
#                Out: ['26835430', '26833718', '26833393', ...]

        if not gene_sym:
            msg = 'ERROR: No gene symbol detected.'
            warnings.warn(msg)
            raise Exception
        if '_' in gene_sym:
            msg = 'ERROR: Gene symbol %s contains an '\
                  'underscore, which is not allowed.' % (gene_sym)
            warnings.warn(msg)
            raise Exception
        query_term = gene_sym + ' [sym]'
        try:
            esearch_records = Entrez.esearch(db='gene', term=query_term,
                                             retmax=retmax, retmod='xml')
        except Exception as e:
            msg = 'ERROR: An error occurred while retrieving '\
                  'data from %s: %s' % ('ESearch', e)
            warnings.warn(msg)
            raise Exception
        parsed_records = Entrez.read(esearch_records)
        entrez_id_list = parsed_records['IdList']
        return entrez_id_list

    @staticmethod
    def _gene_product_lookup(entrez_id_list):
        ''' An internal static function to convert a list of Entrez IDs to a
        list of Entrez gene records via EPost and ESummary.
        Args:
            entrez_id_list (list): a list of Entrez IDs; example: ['26835430',
                                   '26833718', '26833393', ...]
        Returns:
            entrez_rec_list (list): a list of Entrez gene records
        Raises:
            none
        '''

#        Examples:
#            Example 1: # Default behaviour
#                >>> entrez_id_list = ['26835430', '26833718', '26833393']
#                >>> _record_lookup(entrez_id_list)
#                Out: ???

        epost_query = Entrez.epost('gene', id=','.join(entrez_id_list))
        try:
            epost_results = Entrez.read(epost_query)
        except Exception as e:
            msg = 'ERROR: An error occurred while retrieving data from '\
                  '%s: %s' % ('EPost', e)
            warnings.warn(msg)
            raise Exception
        webenv = epost_results['WebEnv']
        query_key = epost_results['QueryKey']
        try:
            esummary_records = Entrez.esummary(db='gene', webenv=webenv,
                                               query_key=query_key)
        except Exception as e:
            msg = 'An error occurred while retrieving data from '\
                  '%s: %s' % ('ESummary', e)
            warnings.warn(msg)
            raise Exception
        entrez_rec_list = Entrez.read(esummary_records)
        return entrez_rec_list

    @staticmethod
    def _parse_gene_products(entrez_rec_list):
        ''' An internal static function to parse out relevant information from .
        Args:
            entrez_rec_list (list): a list of Entrez gene records
        Returns:
            gene_info_list (list): a list of dictionaries
        Raises:
            none
        '''

#        Examples:
#            Example 1: # Default behaviour
#                >>> entrez_rec_list = []
#                >>> _parse_records(entrez_rec_list)

        try:
            documentSummarySet = entrez_rec_list['DocumentSummarySet']
            docs = documentSummarySet['DocumentSummary']
        except Exception as e:
            msg = 'An error occurred while parsing the '\
                  'data from %s: %s' % ('ESummary', e)
            warnings.warn(msg)
            raise Exception
        list_gene_product = [doc['Description'] for doc in docs]
        #list_gene_symbol = [doc['NomenclatureSymbol'] for doc in docs]
        #list_gene_name = [doc['Name'] for doc in docs]
        # Avoiding that spurious first hit biases gene_product:
        gene_product = Counter(list_gene_product).most_common()[0][0]
        return gene_product

    @staticmethod
    def _taxname_lookup_ncbi(taxon_name, retmax=1):
        ''' An internal static function to look up a taxon name at NCBI
            Taxonomy via ESearch.
        Args:
            taxon_name (str): a taxon name; example: 'Pyrus tamamaschjanae'
            retmax (int):     the number of maximally retained hits
        Returns:
            entrez_hitcount (int): an integer
        Raises:
            none
        '''

#        Examples:
#            Example 1: # Default behaviour
#                >>> taxon_name = 'Pyrus tamamaschjanae'
#                >>> _taxname_lookup(taxon_name)
#                Out: 0

        if not taxon_name:
            msg = 'No taxon name detected.'
            warnings.warn(msg)
            raise Exception
        if '_' in taxon_name:
            msg = 'Taxon name %s contains an underscore, '\
                  'which is not allowed.' % (taxon_name)
            warnings.warn(msg)
            raise Exception
        query_term = taxon_name
        try:
            esearch_records = Entrez.esearch(db='taxonomy', term=query_term,
                                             retmax=retmax, retmod='xml')

        except Exception as e:
            msg = 'An error occurred while retrieving data from '\
                  '%s: %s' % ('ESearch', e)
            warnings.warn(msg)
            raise Exception
        parsed_records = Entrez.read(esearch_records)
        entrez_hitcount = parsed_records['Count']
        return str(entrez_hitcount)

    @staticmethod
    def _taxname_lookup_ena(taxon_name, retmax=1):
        ''' An internal static function to look up a taxon name via ENA taxonomy service.
        Args:
            taxon_name (str): a taxon name; example: 'Pyrus tamamaschjanae'
            retmax (int):     the number of maximally retained hits
        Returns:
            entrez_hitcount (int): an integer
        Raises:
            none
        '''

        #        Examples:
        #            Example 1: # Default behaviour
        #                >>> taxon_name = 'Pyrus tamamaschjanae'
        #                >>> _taxname_lookup(taxon_name)
        #                Out: 0

        if not taxon_name:
            msg = 'No taxon name detected.'
            warnings.warn(msg)
            raise Exception
        if '_' in taxon_name:
            msg = 'Taxon name %s contains an underscore, '\
                  'which is not allowed.' % (taxon_name)
            warnings.warn(msg)
            raise Exception
        base_url = "https://www.ebi.ac.uk/ena/data/taxonomy/v1/taxon/scientific-name/"
        final_url = base_url + taxon_name.replace(" ", "%20")
        if os.name == "posix":  # Linux and MacOS
            try:
                enaTaxonomy_records = urlopen(final_url).read()
            except:
                return str(0)
        elif os.name == "nt":  # Windows ## Note: "urlopen" does not work well under Windows, is replaced with "requests"
            try:
                response = requests.get(final_url)
                if response.status_code == 200:
                    return str(1)
                else:
                    raise
            except:
                return str(0)
        entrez_hitcount = enaTaxonomy_records.decode('ascii').count("taxId", 0, len(enaTaxonomy_records.decode('ascii')))
        return str(entrez_hitcount)


    def obtain_gene_product(self, gene_sym):
        ''' This function performs something.
        '''

#        Examples:
#            Example 1: # Default behaviour
#                >>> gene_sym = 'psbI'
#                >>> GetGeneInfo()._entrezid_lookup(gene_sym)
#                Out: ['26835430', '26833718', '26833393', ...]

        Entrez.email = self.email_addr
        try:
            entrez_id_list = GetEntrezInfo._id_lookup(gene_sym)
        except Exception as e:
            print(e)
            raise
        try:
            entrez_rec_list = GetEntrezInfo._gene_product_lookup(
                entrez_id_list)
        except Exception as e:
            print(e)
            raise
        try:
            gene_product = GetEntrezInfo._parse_gene_products(entrez_rec_list)
        except Exception as e:
            print(e)
            raise
        return gene_product


    def does_taxon_exist(self, taxon_name):
        ''' This function calls _taxname_lookup and thus evaluates if a taxon exists.
        Args:
            taxon_name (str): a taxon name; example: 'Pyrus tamamaschjanae'
            retmax (int):     the number of maximally retained hits
        Returns:
            entrez_id_list (list): a list of Entrez IDs; example: ['26835430',
                            '26833718', '26833393', ...]
        Raises:
            none
        '''
        Entrez.email = self.email_addr
        try:
            entrez_hitcount = GetEntrezInfo._taxname_lookup_ena(taxon_name)
            #entrez_hitcount = GetEntrezInfo._taxname_lookup_ncbi(taxon_name)
        except Exception as e:
            print(e)
            raise
        if entrez_hitcount == '0':
            return False
        elif entrez_hitcount == '1':
            return True


class ConfirmAdjustTaxonName:
    ''' This class contains functions to confirm or adjust a sequence's
    taxon name.
    '''

    def __init__(self):
        pass

    def go(self, seq_record, email_addr):
        ''' This function evaluates a taxon name against NCBI taxonomy;
            if not listed, it adjusts the taxon name and appends it
            as ecotype info.
            Args:
                seq_record (obj):   a seqRecord object
                email_addr (dict):  your email address; example:
                                    "m.gruenstaeudl@fu-berlin.de"
            Returns:
                seq_record (obj):   a seqRecord object
            Raises:
                currently nothing
        '''
        try:
            genus_name, specific_epithet = seq_record.name.split(' ', 1)
        except Exception as e:
            msg = 'ERROR: %s: %s:\n.' % ('Could not locate a '\
            'whitespace between genus name and specific epithet in taxon name '\
            'of sequence', seq_record.id)
            warnings.warn(msg)
            raise Exception
        if not GetEntrezInfo(email_addr).does_taxon_exist(seq_record.name):
            msg = 'WARNING: Taxon name of sequence %s '\
                  'not found via ENA taxonomy service: %s. Please consider sending '\
                  'a taxon request via the ENA Webin interface.'\
                   % (seq_record.id, seq_record.name)
            warnings.warn(msg)
            if not GetEntrezInfo(email_addr).does_taxon_exist(genus_name):
                msg = 'ERROR: Neither genus name, '\
                      'nor species name of sequence %s were found in '\
                      'via ENA taxonomy service' % (seq_record.id)
                warnings.warn(msg)
                raise Exception
            else:
                species_name_original = seq_record.name
                species_name_new = genus_name + ' sp. ' + specific_epithet
                seq_record.name = species_name_new
                seq_record.features[0].qualifiers['organism'] = species_name_new
                seq_record.description = seq_record.description.\
                    replace(species_name_original, species_name_new)
                msg = 'WARNING: Taxon name of sequence '\
                      '%s converted to the informal name: %s'\
                      % (seq_record.id, species_name_new)
                warnings.warn(msg)
        return seq_record


class ParseCharsetName:
    ''' This class contains functions to parse charset names.
    Args:
        charset_name (str): a string that represents a charset name; example:
                            "psbI_CDS"
        email_addr (dict):  your email address; example:
                            "your_email_here@yourmailserver.com"
        product_lookup (bool): decision if product name shall be looked up
    Raises:
        currently nothing
    '''

    def __init__(self, charset_name, email_addr, product_lookup):
        self.charset_name = charset_name
        self.email_addr = email_addr
        self.product_lookup = product_lookup

    @staticmethod
    def _extract_charset_info(charset_name):
        charset_orient = False
        charset_type = False
        charset_sym = False
        orient_present = [ori for ori in GlobVars.nex2ena_valid_orientations if ori in charset_name]
        try:
            if(len(orient_present) == 0):
                 charset_orient = 'forw'
            elif(len(orient_present) == 1):
                charset_orient = orient_present[0]
                if charset_orient == "forw":
                    charset_name = charset_name.replace("forward","")
                    charset_name = charset_name.replace("forw","")
                elif charset_orient == "rev":
                    charset_name = charset_name.replace("reverse","")
                    charset_name = charset_name.replace("rev","")
        except Exception as e:
            msg = 'ERROR: %s:\n %s' % ('Unclear parsing of feature orientation', e)
            warnings.warn(msg)
            raise Exception
        type_present = [typ for typ in GlobVars.nex2ena_valid_INSDC_featurekeys if typ in charset_name]
        try:
            if(len(type_present) == 0):
                msg = 'ERROR: %s:\n %s' % ('No valid feature keys', e)
                warnings.warn(msg)
                raise Exception
            elif(len(type_present) == 1):
                charset_type = type_present[0]
                charset_name = ''.join(charset_name.split(type_present[0]))
            elif(len(type_present) > 1):
                msg = 'WARNING: More than one charset_type '\
                'encountered in charset: %s' % (charset_name)
                warnings.warn(msg)
                charset_type = type_present[0]
                charset_name = ''.join(charset_name.split(type_present[0]))
        except Exception as e:
            msg = 'ERROR: %s:\n %s' % ('Unclear parsing of features', e)
            warnings.warn(msg)
            raise Exception
        charset_sym = charset_name.strip('_').split('_')
        try:
            if len(charset_sym) == 1:
                return (charset_sym[0], charset_type, charset_orient)
        except Exception as e:
            msg = 'ERROR: %s:\n %s' % ('Unspecified error '\
            'during feature parsing', e)
            warnings.warn(msg)
            raise Exception


    def parse(self):
        ''' This function parses the charset_name.
        Returns:
            tupl.   The return consists of three strings in the order
                    "charset_sym, charset_type, charset_orient, charset_product"
        '''
        try:
            charset_sym, charset_type, charset_orient = ParseCharsetName._extract_charset_info(self.charset_name)
        except Exception as e:
            msg = 'ERROR: %s:\n %s' % ('Error while parsing the charset_name', e)
            warnings.warn(msg)
            raise Exception
        entrez_handle = GetEntrezInfo(self.email_addr)
        if (charset_type == 'CDS' or charset_type == 'gene') and self.product_lookup:
            try:
                charset_product = entrez_handle.obtain_gene_product(charset_sym)
            except Exception as e:
                msg = 'ERROR: %s:\n %s' % ('Error while obtaining gene product', e)
                warnings.warn(msg)
                raise Exception
        else:
            charset_product = None
        return (charset_sym, charset_type, charset_orient, charset_product)

class SequenceParsing:
    '''
    This class contains functions to parse and modify DNA sequence strings.
    '''
    def __init__(self):
        pass

    def reverseComplement(sequence):
        ''' This function reverse complements a DNA sequence string
            Args:
                sequence (str)
            Returns:
                reverse complemented sequence (str)
            Raises:
                currently nothing
        '''

        reverseComplement = []
        for nucl in sequence:
            if nucl == "A":
                reverseComplement.append("T")
            elif nucl == "T":
                reverseComplement.append("A")
            elif nucl == "C":
                reverseComplement.append("G")
            elif nucl == "G":
                reverseComplement.append("C")
        return ''.join(reverseComplement)[::-1]



#!/usr/bin/env python
'''
Custom operations to check annotations
'''

#####################
# IMPORT OPERATIONS #
#####################

import GenerationOps as GnOps
import GlobalVariables as GlobVars
import warnings

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import FeatureLocation
from Bio.SeqFeature import CompoundLocation
from unidecode import unidecode
from itertools import chain

###############
# AUTHOR INFO #
###############

__author__ = 'Michael Gruenstaeudl <m.gruenstaeudl@fu-berlin.de>'
__copyright__ = 'Copyright (C) 2016-2019 Michael Gruenstaeudl'
__info__ = 'annonex2embl'
__version__ = '2019.10.16.1700'

#############
# DEBUGGING #
#############

#import ipdb
#ipdb.set_trace()

###########
# CLASSES #
###########

class AnnoCheck:
    ''' This class contains functions to evaluate the quality of an
        annotation.
    Args:
        extract (obj):      a sequence object; example: Seq('ATGGAGTAA',
                            IUPACAmbiguousDNA())
        feature_object (obj):   a feature object
        record_id (str):    a string deatiling the name of the sequence in
                            question; example: "taxon_A"
        transl_table (int): an integer; example: 11 (for bacterial code)
    Returns:
        tupl.   The return consists of the translated sequence (a str)
                and the updated feature location (a location object);
                example: (transl_out, feat_loc)
    Raises:
        Exception
    '''

    def __init__(self, extract, feature, record_id, transl_table=11):
        self.extract = extract
        self.feature = feature
        self.record_id = record_id
        self.transl_table = transl_table

    @staticmethod
    def _transl(extract, transl_table, to_stop=False, cds=False):
        ''' An internal static function to translate a coding region. '''
        
        with warnings.catch_warnings():  # Note: Suppressing warnings necessary to suppress the Biopython warning about an annotation not being a multiple of three
            warnings.filterwarnings("ignore")
            transl = extract.translate(table=transl_table, to_stop=to_stop,
                                       cds=cds)
            # Adjustment for non-start codons given the necessary use of
            # cds=True in TPL.
            if not extract.startswith(GlobVars.nex2ena_start_codon):
                first_codon_seq = extract[0:3]
                first_aa = first_codon_seq.translate(table=transl_table,
                                                     to_stop=to_stop, cds=False)
                transl = first_aa + transl[1:]
        return transl

    @staticmethod
    def _check_protein_start(extract, transl_table):
        ''' An internal static function to translate a coding region and check
            if it starts with a methionine. '''
        transl = extract.translate(table=transl_table)
        return transl.startswith('M')

    @staticmethod
    def _adjust_feat_loc(location_object, transl_with_internStop,
                         transl_without_internStop):
        ''' An internal static function to adjust the feature location if an
            internal stop codon were present. '''
        if len(transl_without_internStop) > len(transl_with_internStop):
            # 1. Unnest the nested lists
            contiguous_subsets = [list(range(e.start.position, e.end.position))
                for e in location_object.parts]
            compound_integer_range = sum(contiguous_subsets, [])
            # 2. Adjust location range
            len_with_internStop = len(transl_with_internStop) * 3
            # IMPORTANT!: In TFL, the "+3" is for the stop codon, which is
            # counted in the location range, but is not part of the AA
            # sequence of the translation.
            adjusted_range = compound_integer_range[:(len_with_internStop+3)]
            # 3. Establish location
            feat_loc = GnOps.GenerateFeatLoc().make_location(adjusted_range)
        if len(transl_without_internStop) == len(transl_with_internStop):
            feat_loc = location_object
        return feat_loc

    def check(self):
        ''' This function performs checks on a coding region.
            Specifically, the function tries to translate the coding
            region (CDS) directly, using the internal checker
            "cds=True". If a direct translation fails, it confirms if
            the CDS starts with a methionine. If the CDS does not start
            with a methionine, a ValueError is raised. If the CDS does
            start with a methionine, translations are conducted with
            and without regard to internal stop codons. The shorter
            of the two translations is kept. The feature location is
            adjusted, where necessary.
        Note:
            The asterisk indicating a stop codon is truncated under
            _transl(to_stop=True) and must consequently be added again
            (see line 137).
        '''
        try:
            # Note: TFL must contain "cds=True"; don't delete it
            transl_out = AnnoCheck._transl(self.extract,
                                           self.transl_table, cds=True)
            feat_loc = self.feature.location
        except:
            try:
                without_internalStop = AnnoCheck._transl(self.extract,
                                                         self.transl_table)
                with_internalStop = AnnoCheck._transl(
                    self.extract, self.transl_table, to_stop=True)
                transl_out = with_internalStop
                feat_loc = AnnoCheck._adjust_feat_loc(
                    self.feature.location, with_internalStop, without_internalStop)
            except Exception:
                msg = 'Translation of feature `%s` of '\
                      'sequence `%s` is unsuccessful.' % (self.feature.id, self.record_id)
                #warnings.warn(msg)
                raise Exception(msg)
        if len(transl_out) < 2:
            msg = 'Translation of feature `%s` of '\
                  'sequence `%s` indicates a protein length of only a '\
                  'single amino acid.' % (self.feature.id, self.record_id)
            #warnings.warn(msg)
            raise Exception(msg)
        # IMPORTANT!!!: In an ENA record, the translation does not display the
        # stop codon (i.e., the '*'), while the feature location range (i.e., 738..2291)
        # very much includes its position, which is biologically logical, as
        # a stop codon is not an amino acid in a translation.
        # Thus, TFL would be incorrect, because it would add back an asterisk into the translation.
        #transl_out = transl_out + "*"
        return (transl_out, feat_loc)

    def for_unittest(self):
        try:
            transl_out, feat_loc = AnnoCheck(
                self.extract, self.feature, self.record_id, self.transl_table).check()
            if isinstance(transl_out, Seq) and isinstance(feat_loc,
                                                          FeatureLocation):
                return True
            return False
        # except ValueError: # Keep 'ValueError'; don't replace with 'Exception'
        #    return False
        except Exception as e:
            print(e)
            raise


class TranslCheck:
    ''' This class contains functions to coordinate different checks. '''

    def __init__(self):
        pass

    # This function extract the sequence from pattern sequence
    # for a forward and reverse strand
    def extract(self, feature, seq_record):
        if feature._get_strand() == 1:
            return feature.extract(seq_record)
        else:
            reverse = SeqRecord("")
            for i in feature.location.parts[::-1]:
                reverse.seq = reverse.seq + i.extract(seq_record).seq
            return reverse

    # By checking the translation of a CDS or an gene it may happen that
    # the location from the CDS or gene had to be adjusted. If after such
    # a feature a IGS or intron follows it have to be adjust aswell.
    # This is done by this function
    def adjustLocation(self, oldLocation, newLocation):
        start = []
        end = []
        start.append(newLocation.end)
        t = oldLocation.start
        for i in oldLocation:
            if not i == t:
                end.append(t)
                start.append(i)
                t = i
            t = t + 1
        end.append(oldLocation.end)
        locations = []
        for i in range(len(start)):
            locations.append(FeatureLocation(start[i],end[i]))
        try:
            return CompoundLocation(locations)
        except Exception as e:
            return locations[0]

    def transl_and_quality_of_transl(self, seq_record, feature, transl_table):
        ''' This function conducts a translation of a coding region and checks
            the quality of said translation.
        Args:
            seq_record (obj):   foobar; example: 'foobar'
            feature (obj):      foobar; example: 'foobar'
            transl_table (int):
        Returns:
            True, unless exception
        Raises:
            feature
        '''
        extract = self.extract(feature, seq_record)
        try:
            transl, loc = AnnoCheck(extract.seq, feature, seq_record.id,
                                    transl_table).check()
            if feature.type == 'CDS':
                feature.qualifiers["translation"] = transl
            if feature.type == 'exon' or feature.type == 'gene':
                # With gene and exon features that are less than 15 nt long,
                # the annotation should be dropped from the output.
                if len([base for base in loc]) < 15:
                    raise
            feature.location = loc
        except Exception as e:
            raise Exception(e)
        return feature


class QualifierCheck:
    ''' This class contains functions to evaluate the quality of metadata.
    Args:
        lst_of_dcts (list): a list of dictionaries; example:
                            [{'foo': 'foobarqux', 'bar': 'foobarqux',
                              'qux': 'foobarqux'}, {'foo': 'foobarbaz',
                              'bar': 'foobarbaz', 'baz': 'foobarbaz'}]
        label (???): ?
    Returns:
        none
    Raises:
        Exception
    '''

    def __init__(self, lst_of_dcts, label):
        self.lst_of_dcts = lst_of_dcts
        self.label = label

    @staticmethod
    def _enforce_ASCII(lst_of_dcts):
        ''' This function converts any non-ASCII characters among
            qualifier values to ASCII characters. '''
        try:
            filtered_lst_of_dcts = [
                {k: unidecode(v) for k, v in list(dct.items())}
                for dct in lst_of_dcts]
        except:
            filtered_lst_of_dcts = [
                {k: unidecode(v.decode('utf-8')) for k, v in dct.items()}
                for dct in lst_of_dcts]
        return filtered_lst_of_dcts

    @staticmethod
    def _label_present(lst_of_dcts, label):
        ''' This function checks if each (!) list of dictionary keys
            of a list of dictionaries encompass the element <label> at
            least once. '''
        if not all(label in list(dct.keys()) for dct in lst_of_dcts):
            msg = 'ERROR: csv-file does not '\
                  'contain a column labelled %s' % (label)
            warnings.warn(msg)
            raise Exception
        return True

    @staticmethod
    def _rm_empty_qual(lst_of_dcts):
        ''' This function removes any qualifier from a dictionary which
            displays an empty value. Technically, this function
            loops through the qualifier dictionaries and removes any
            key-value-pair from a dictionary which contains an empty
            value. '''
        nonempty_lst_of_dcts = [{k: v for k, v in list(dct.items()) if v != ''}
                                for dct in lst_of_dcts]
        return nonempty_lst_of_dcts

    @staticmethod
    def _valid_INSDC_quals(lst_of_dcts):
        ''' This function checks if every (!) dictionary key in a list of
            dictionaries is a valid INSDC qualifier. '''
        keys_present = list(chain.from_iterable([list(dct.keys()) for dct in
                                                 lst_of_dcts]))
        not_valid = [k for k in keys_present if k not in
                     GlobVars.nex2ena_valid_INSDC_quals]
        if not_valid:
            msg = 'ERROR: The following are '\
                  'invalid INSDC qualifiers: %s' % (', '.join(not_valid))
            warnings.warn(msg)
            raise Exception
        return True

    def quality_of_qualifiers(self):
        ''' This function conducts a series of quality checks on the
            qualifiers list (a list of dictionaries). First (label_present),
            it checks if a qualifier matrix (and, hence, each entry)
            contains a column labelled with <seqname_col_label>.
            Second (nex2ena_valid_INSDC_quals), it checks if column
            names constitute valid INSDC feature table qualifiers.
        Args:
            label (str):  a string; example: 'isolate'
            lst_of_dcts (list): a list of dictionaries; example:
                                [{'isolate': 'taxon_A', 'country': 'Ecuador'},
                                 {'isolate': 'taxon_B', 'country': 'Peru'}]
        Returns:
            True, unless exception
        Raises:
            passed exception
        '''
        try:
            QualifierCheck._label_present(self.lst_of_dcts, self.label)
        except Exception as e:
            warnings.warn(e)
            raise Exception
        try:
            QualifierCheck._valid_INSDC_quals(self.lst_of_dcts)
        except Exception as e:
            warnings.warn(e)
            raise Exception
        return True



#!/usr/bin/env python2.7

''' Main operations in annonex2embl '''

#####################
# IMPORT OPERATIONS #
#####################

import CheckingOps as CkOps
import DegappingOps as DgOps
import GenerationOps as GnOps
import GlobalVariables as GlobVars
import ParsingOps as PrOps
import IOOps as IOOps
import sys
import os
import gzip
import warnings

from Bio import SeqIO
#from Bio.Alphabet import generic_dna
#from Bio.Seq import Seq
from Bio import SeqFeature
from collections import OrderedDict
from copy import copy
from copy import deepcopy
from distutils.util import strtobool
from unidecode import unidecode

# Add specific directory to sys.path in order to import its modules
# Note: This relative importing is amateurish. Could the following
#       import be replaced with 'import annonex2embl'?
sys.path.append(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'annonex2embl'))

###############
# AUTHOR INFO #
###############

__author__ = 'Michael Gruenstaeudl <m.gruenstaeudl@fu-berlin.de>'
__copyright__ = 'Copyright (C) 2016-2019 Michael Gruenstaeudl'
__info__ = 'annonex2embl'
__version__ = '2019.10.16.1700'

#############
# DEBUGGING #
#############

#import ipdb
#ipdb.set_trace()

# To format warnings in a pretty, readable way:
def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return '\n annonex2embl %s\n' % (message)
warnings.formatwarning = warning_on_one_line

###########
# CLASSES #
###########

#############
# FUNCTIONS #
#############


warnings.formatwarning = warning_on_one_line

def annonex2embl(path_to_nex,
                 path_to_csv,
                 descr_DEline,
                 email_addr,
                 author_names,
                 path_to_outfile,

                 manifest_study='',
                 manifest_descr='',
                 product_lookup='False',
                 tax_check='False',
                 linemask='False',
                 topology='linear',
                 tax_division='PLN',
                 uniq_seqid_col='isolate',
                 transl_table='11',
                 organelle='plastid',
                 seq_version='1',
                 compress='False'):

########################################################################

# 0. MAKE SPECIFIC VARIABLES BOOLEAN
    productlookup_bool = strtobool(product_lookup)
    taxcheck_bool = strtobool(tax_check)
    linemask_bool = strtobool(linemask)
    compress_bool = strtobool(compress)

########################################################################

# 1. OPEN OUTFILE
    with open(path_to_outfile, 'a') as outp_handle:

########################################################################

# 2. PARSE DATA FROM .NEX-FILE
        try:
            charsets_global, alignm_global = IOOps.Inp().\
                parse_nexus_file(path_to_nex)
        except Exception as e:
            msg = 'ERROR: %s' % (e)
            warnings.warn(msg)
            raise Exception

########################################################################

# 3. PARSE DATA FROM .CSV-FILE
        try:
            raw_qualifiers = IOOps.Inp().parse_csv_file(path_to_csv)
        except Exception as e:
            msg = 'ERROR: %s' % (e)
            warnings.warn(msg)
            raise Exception

########################################################################

# 4.1 CHECK QUALIFIERS
# 4.1.1 Perform quality checks on qualifiers
        try:
            CkOps.QualifierCheck(raw_qualifiers, uniq_seqid_col).\
                quality_of_qualifiers()
        except Exception as e:
            msg = 'ERROR: %s' % (e)
            warnings.warn(msg)
            raise Exception
    # 4.1.2 Remove qualifiers without content (i.e. empty qualifiers)
        nonempty_qualifiers = CkOps.QualifierCheck.\
            _rm_empty_qual(raw_qualifiers)
    # 4.1.3 Enforce that all qualifier values consist of ASCII characters
        filtered_qualifiers = CkOps.QualifierCheck.\
            _enforce_ASCII(nonempty_qualifiers)

####################################

# 4.2 CHECK SEQUENCES
        sorted_seqnames = sorted(alignm_global.keys())
        sorted_seqids = sorted([d[uniq_seqid_col] for d in filtered_qualifiers])
    # 4.2.1. Exit if seq names in NEX-file not identical to seq ids in csv-file
        try:
            not_shared = list(set(sorted_seqnames) - set(sorted_seqids))
        except Exception as e:
            msg = 'ERROR: Sequence names in %s are NOT IDENTICAL to sequence IDs in %s.'\
                  '%s The following sequence names don\'t have a match: %s'\
                  % (path_to_nex, path_to_csv, '\n', ','.join(not_shared))
            warnings.warn(msg)
            raise Exception

########################################################################

# 5. PARSE OUT FEATURE KEY, OBTAIN OFFICIAL GENE NAME AND GENE PRODUCT
        charset_dict = {}
        for charset_name in list(charsets_global.keys()):
            try:
                charset_sym, charset_type, charset_orient, charset_product = PrOps.\
                    ParseCharsetName(charset_name, email_addr, productlookup_bool).parse()
            except Exception as e:
                msg = 'ERROR: %s' % (e)
                warnings.warn(msg)
                raise Exception
            charset_dict[charset_name] = (charset_sym, charset_type, charset_orient,
                                          charset_product)

########################################################################

# 6. GENERATING SEQ_RECORDS BY LOOPING THROUGH EACH SEQUENCE OF THE ALIGNMENT
#    Work off the sequences alphabetically.
        for counter, seq_name in enumerate(sorted_seqnames):
            # TFLs generate safe copies of charset and alignment for every loop iteration; however, the cmd "copy()" itself would not work because it cannot operate on lists of lists; thus, we use "deepcopy()", which can.
            charsets_withgaps = deepcopy(charsets_global)
            alignm = deepcopy(alignm_global)

####################################

# 6.1. SELECT CURRENT SEQUENCES AND CURRENT QUALIFIERS
            current_seq = alignm[seq_name]
            current_quals = [d for d in filtered_qualifiers
                             if d[uniq_seqid_col] == seq_name][0]

####################################

# 6.2. GENERATE THE BASIC SEQ_RECORD (I.E., WITHOUT FEATURES)

# 6.2.1. Generate the basic SeqRecord
            seq_record = GnOps.GenerateSeqRecord().base_record(
                current_seq, current_quals, uniq_seqid_col, seq_version,
                descr_DEline, topology, tax_division, organelle)
            # Add a function that automatically removes all sequences
            # that consist only of Ns (or ?s).
            skip = True
            for i in seq_record.seq:
                if i != 'N' and i != '?':
                    skip = False
                    break
            if skip:
                continue

####################################

# 6.3. CLEAN UP THE SEQUENCE OF THE SEQ_RECORD (i.e., remove leading or
#      trailing ambiguities, remove gaps), but maintain correct
#      annotations.
#      Note 1: This clean-up has to occur before (!) the SeqFeature
#      'source' is generated, as the source feature provides info on
#      the full sequence length.
#      Note 2: Charsets are identical across all sequences.

# 6.3.1. Replace question marks in DNA sequence with 'N'
            seq_record.seq._data = seq_record.seq._data.replace('?', 'N')
            # TFL generates a safe copy of sequence to work on
            seq_withgaps = copy(seq_record.seq)

# 6.3.2. Skip all sequence records smaller than 10 unambiguous nucleotides
            if len(seq_record.seq._data.replace('-','').strip('N')) <= 10:
                msg = 'WARNING: Sequence `%s` not saved because shorter than 10 '\
                      'unambiguous nucleotides.' % (seq_record.id)
                warnings.warn(msg)
                continue

# 6.3.3. Remove leading ambiguities while maintaining correct annotations
            seq_noleadambigs, charsets_noleadambigs = DgOps.\
                RmAmbigsButMaintainAnno().rm_leadambig(seq_withgaps, 'N',
                                                       charsets_withgaps)

# 6.3.4. Remove trailing ambiguities while maintaining correct annotations
            seq_notrailambigs, charsets_notrailambigs = DgOps.\
                RmAmbigsButMaintainAnno().rm_trailambig(seq_noleadambigs,
                                                        'N', charsets_noleadambigs)
# 6.3.5. Degap the sequence while maintaining correct annotations
            seq_nogaps, charsets_degapped = DgOps.\
                DegapButMaintainAnno(seq_notrailambigs, '-',
                                     charsets_notrailambigs).degap()

# 6.3.6. Add gap features where stretches of Ns in sequence
            seq_final, charsets_final = DgOps.\
                AddGapFeature(seq_nogaps, charsets_degapped).add()
            # TFL assigns the deambiged and degapped sequence back
            seq_record.seq = seq_final

####################################

# 6.4. GENERATE SEQFEATURE 'SOURCE' AND TEST TAXON NAME AGAINST
#      NCBI TAXONOMY

# 6.4.1. Generate SeqFeature 'source' and append to features list
            charset_names = list(charsets_final.keys())
            source_feature = GnOps.GenerateSeqFeature().\
                source_feat(len(seq_record), current_quals, charset_names)
            seq_record.features.append(source_feature)

####################################

# 6.5. VALIDATE TAXON NAME

# 6.5.1. Test taxon name against NCBI taxonomy; if not listed, adjust
#        taxon name and append ecotype info
            if taxcheck_bool:
                seq_record = PrOps.ConfirmAdjustTaxonName().go(seq_record,
                                                               email_addr)

####################################

# 6.6. POPULATE THE FEATURE KEYS WITH THE CHARSET INFORMATION
#      Note: Each charset represents a dictionary that must be added in
#      full to the list "SeqRecord.features"

            for charset_name, charset_range in list(charsets_final.items()):

# 6.6.1. Proceed in loop only if charset_range is not empty
#        An empty charset_range could be the case if the charset only
#        consisted of 'N' (which were removed in steps 6.3.2 and 6.3.3).
                if charset_range:
# 6.6.2. Convert charset_range into Location Object
                    location_object = GnOps.GenerateFeatLoc().make_location(charset_range)

# 6.6.3. Assign a gene product to a gene name, unless it's a gap feature
                    if charset_name[0:3] == "gap":
                        charset_sym = None
                        charset_type = "gap"
                        charset_orient = "forw"
                        charset_product = None
                    else:
                        charset_sym, charset_type, charset_orient, charset_product = charset_dict[charset_name]

# 6.6.4. Generate a regular SeqFeature and append to seq_record.features
#        Note: The position indices for the stop codon are truncated in
#              this step.
                    seq = []
                    [seq.append(seq_record[obj]) for obj in location_object]
                    seq = ''.join(seq)

                    seq_feature = GnOps.GenerateSeqFeature().regular_feat(
                        charset_sym, charset_type, charset_orient, location_object, transl_table,
                        seq, charset_product)
                    seq_record.features.append(seq_feature)

####################################

# 6.7. SORT ALL SEQ_RECORD.FEATURES EXCEPT THE FIRST ONE (WHICH
#      CONSTITUTES THE SOURCE FEATURE) BY THEIR RELATIVE START
#      POSITIONS
            sorted_features = sorted(seq_record.features[1:],
                                     key=lambda x: x.location.start.position)
            seq_record.features = [seq_record.features[0]] + sorted_features

####################################

# 6.8. TRANSLATE AND CHECK QUALITY OF TRANSLATION
            removal_list = []
            last_seen = ["type", "before", "after"]
            for indx, feature in enumerate(seq_record.features):
                # Check if feature is a coding region
                if feature.type == 'CDS' or feature.type == 'gene':
                    try:
                        # In TFL, features are truncated to the first
                        # internal stop codon, if present.
                        last_seen[0] = feature.type
                        last_seen[1] = feature.location
                        feature = CkOps.TranslCheck().\
                            transl_and_quality_of_transl(seq_record,
                                                         feature, transl_table)
                        last_seen[2] = feature.location
                    except Exception as e:
                        msg = 'WARNING: Feature `%s` (type: %s) of sequence `%s` is not saved to '\
                              'output. Reason: %s' % (feature.id, feature.type, seq_record.id, e)
                        warnings.warn(msg)
                        removal_list.append(indx)
                elif feature.type == 'IGS' or feature.type == 'intron':
                    if  last_seen[0] == 'CDS' or last_seen[0] == 'gene':
                        if not last_seen[1] == last_seen[2]:
                            feature.location = CkOps.TranslCheck().\
                                                    adjustLocation(feature.location, last_seen[2])
                    last_seen = ["type","loc_before","loc_after"]
                else:
                    last_seen = ["type","loc_before","loc_after"]
            # TFL removes the objects in reverse order, because otherwise
            # each removal would shift the indices of subsequent objects
            # to the left.
            for indx in sorted(removal_list, reverse=True):
                seq_record.features.pop(indx)

####################################

# 6.9. INTRODUCE FUZZY ENDS
            for feature in seq_record.features:
                # Check if feature is a coding region
                if feature.type == 'CDS' or feature.type == 'gene':
                    # Note: Don't use "feature.extract(seq_record.seq)" in TFLs,
                    #       as stop codon was truncated from feature under
                    #       Step 6.8, because in an ENA record, the AA sequence
                    #       of the translation does not have the stop codon
                    #       (i.e., the '*'), while the feature location
                    #       range (i.e., 738..2291) very much includes
                    #       its position (which is biologically logical).
                    charset_range_updated = list(range(feature.location.start.position,
                        feature.location.end.position))
                    coding_seq = ''.join([seq_record.seq[i] for i in charset_range_updated])
                    if feature.location._get_strand() == -1:
                        coding_seq = PrOps.SequenceParsing.reverseComplement(coding_seq)
                        feature.location.parts = feature.location.parts[::-1] # if feature reverse, the feature components list must be inverted
                    if not coding_seq.startswith(GlobVars.nex2ena_start_codon):
                        feature.location = GnOps.GenerateFeatLoc().\
                            make_start_fuzzy(feature.location)
                    if all([not coding_seq.endswith(c)
                            for c in GlobVars.nex2ena_stop_codons]):
                                feature.location = GnOps.GenerateFeatLoc().\
                                make_end_fuzzy(feature.location)


# (FUTURE)  Also introduce fuzzy ends to features when those had leading or trailing Ns removed,
#           because the removed Ns may constitute start of stop codons.

####################################

# 6.10. DECISION ON OUTPUT FORMAT
            IOOps.Outp().write_SeqRecord(seq_name, seq_record,
                                         author_names, outp_handle,
                                         linemask_bool)

########################################################################

# 7. CREATE COMPRESSED FILE

    if compress_bool:
        with open(path_to_outfile, 'r') as outp_handle:
            lines = outp_handle.readlines()
        with gzip.open(path_to_outfile + '.gz', 'wb') as outp_handle:
            for line in lines:
                try:
                    outp_handle.write(bytearray(line, 'utf-8'))
                except Exception as e:
                    try:
                        outp_handle.write(line)
                    except Exception as e:
                        msg = 'WARNING: EMBL flatfile not compressed '\
                              'due to error: %s' % (e)
                        warnings.warn(msg)

########################################################################

# 8. CREATE MANIFEST FILE

    if manifest_study and manifest_descr:
        manifest_flatfile = os.path.basename(path_to_outfile)
        if compress_bool:
            IOOps.Outp().create_manifest(path_to_outfile,
                                        manifest_study,
                                        manifest_descr,
                                        manifest_flatfile+'.gz')
        else:
            IOOps.Outp().create_manifest(path_to_outfile,
                                        manifest_study,
                                        manifest_descr,
                                        manifest_flatfile)

    elif manifest_study and not manifest_descr:
        msg = 'WARNING: Manifest file not written '\
              'due to missing manifest name.'
        warnings.warn(msg)

    elif not manifest_study and manifest_descr:
        msg = 'WARNING: Manifest file not written '\
              'due to missing manifest study.'
        warnings.warn(msg)
    else:
        pass

########################################################################



#!/usr/bin/env python
'''
Setting global variables.
'''

###############
# AUTHOR INFO #
###############

__author__ = 'Michael Gruenstaeudl <m.gruenstaeudl@fu-berlin.de>'
__copyright__ = 'Copyright (C) 2016-2019 Michael Gruenstaeudl'
__info__ = 'annonex2embl'
__version__ = '2019.10.16.1700'

#############
# DEBUGGING #
#############

#import ipdb
#ipdb.set_trace()

#########
# OTHER #
#########

# Valid start codon as defined by [...]
# http://
#global nex2ena_start_codon
nex2ena_start_codon = 'ATG'

# Valid stop codons as defined by [...]
# http://
#global nex2ena_stop_codons
nex2ena_stop_codons = ['TAG', 'TAA', 'TGA']  # amber, ochre, opal

# Valid sequence topology as defined by the EMBL User Manual
# (Chapter 3.4.1).
# ftp://ftp.ebi.ac.uk/pub/databases/embl/doc/usrman.txt
#global nex2ena_valid_topologies
nex2ena_valid_topologies = ['linear', 'circular']

# Valid taxonomic divisions as defined by the EMBL User Manual
# (Chapter 3.2).
# ftp://ftp.ebi.ac.uk/pub/databases/embl/doc/usrman.txt
#global nex2ena_valid_tax_divisions
nex2ena_valid_tax_divisions = [
    'PHG', 'ENV', 'FUN', 'HUM', 'INV', 'MAM', 'VRT', 'MUS',
    'PLN', 'PRO', 'ROD', 'SYN', 'TGN', 'UNC', 'VRL', 'XXX']
#    See Section '3.2 Taxonomic Division' in EMBL User Manual
#        Division                 Code
#        -----------------        ----
#        Bacteriophage            PHG
#        Environmental Sample     ENV
#        Fungal                   FUN
#        Human                    HUM
#        Invertebrate             INV
#        Other Mammal             MAM
#        Other Vertebrate         VRT
#        Mus musculus             MUS
#        Plant                    PLN
#        Prokaryote               PRO
#        Other Rodent             ROD
#        Synthetic                SYN
#        Transgenic               TGN
#        Unclassified             UNC (i.e. unknown)
#        Viral                    VRL
#        ENA SUBMISSIONS          XXX


# Valid feature table qualifiers as defined by the International
# Nucleotide Sequence Database Collection (INSDC)
# http://www.insdc.org/files/feature_table.html#7.3.1
#global nex2ena_valid_INSDC_quals
nex2ena_valid_INSDC_quals = [
    'allele', 'altitude', 'anticodon', 'artificial_location',
    'bio_material', 'bound_moiety', 'cell_line', 'cell_type',
    'chromosome', 'citation', 'clone', 'clone_lib', 'codon_start',
    'collected_by', 'collection_date', 'compare', 'country',
    'cultivar', 'culture_collection', 'db_xref', 'dev_stage',
    'direction', 'EC_number', 'ecotype', 'environmental_sample',
    'estimated_length', 'exception', 'experiment', 'focus',
    'frequency', 'function', 'gap_type', 'gene', 'gene_synonym',
    'germline', 'haplogroup', 'haplotype', 'host', 'identified_by',
    'inference', 'isolate', 'isolation_source', 'lab_host',
    'lat_lon', 'linkage_evidence', 'locus_tag', 'macronuclear',
    'map', 'mating_type', 'mobile_element_type', 'mod_base',
    'mol_type', 'ncRNA_class', 'note', 'number', 'old_locus_tag',
    'operon', 'organelle', 'organism', 'partial', 'PCR_conditions',
    'PCR_primers', 'phenotype', 'plasmid', 'pop_variant', 'product',
    'protein_id', 'proviral', 'pseudo', 'pseudogene', 'rearranged',
    'regulatory_class', 'replace', 'ribosomal_slippage',
    'rpt_family', 'rpt_type', 'rpt_unit_range', 'rpt_unit_seq',
    'satellite', 'segment', 'serotype', 'serovar', 'sex',
    'specimen_voucher', 'standard_name', 'strain', 'sub_clone',
    'sub_species', 'sub_strain', 'tag_peptide', 'tissue_lib',
    'tissue_type', 'transgenic', 'translation', 'transl_except',
    'transl_table', 'trans_splicing', 'type_material', 'variety']

# Valid feature table qualifiers as defined by the International
# Nucleotide Sequence Database Collection (INSDC)
# http://www.insdc.org/files/feature_table.html#7.3.1
#global nex2ena_valid_INSDC_featurekeys
nex2ena_valid_INSDC_featurekeys = [
    'assembly_gap', 'C_region', 'CDS', 'centromere',
    'D-loop', 'D_segment', 'exon', 'gap', 'gene', 'iDNA',
    'intron', 'J_segment', 'LTR', 'mat_peptide', 'misc_binding',
    'misc_difference', 'misc_feature', 'misc_recomb', 'misc_RNA',
    'misc_structure', 'mobile_element', 'modified_base', 'mRNA',
    'ncRNA', 'N_region', 'old_sequence', 'operon', 'oriT',
    'polyA_site', 'precursor_RNA', 'prim_transcript', 'primer_bind',
    'protein_bind', 'regulatory', 'repeat_region', 'rep_origin',
    'rRNA', 'S_region', 'sig_peptide', 'source', 'stem_loop',
    'STS', 'telomere', 'tmRNA', 'transit_peptide', 'tRNA',
    'unsure', 'V_region', 'V_segment', 'variation', "3'UTR", "5'UTR",
    "IGS"]

# Valid organelle qualifiers as defined by the International
# Nucleotide Sequence Database Collection (INSDC)
# http://www.insdc.org/files/feature_table.html#7.3.1
#global nex2ena_valid_INSDC_organelle
nex2ena_valid_INSDC_organelle = [
    'chromatophore', 'hydrogenosome', 'mitochondrion', 'nucleomorph',
    'plastid', 'mitochondrion:kinetoplast', 'plastid:chloroplast',
    'plastid:apicoplast', 'plastid:chromoplast', 'plastid:cyanelle',
    'plastid:leucoplast', 'plastid:proplastid']

nex2ena_valid_orientations = [
    'forw','rev']



import os
import glob
import unittest
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def my_test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('tests', pattern='*_test.py')
    return test_suite

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='annonex2embl',
    version='0.8.0',
    author='Michael Gruenstaeudl, PhD',
    author_email='m.gruenstaeudl@fu-berlin.de',
    description='Converts an annotated DNA multi-sequence alignment (in NEXUS format) to an EMBL flatfile for submission to ENA via the Webin-CLI submission tool',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/michaelgruenstaeudl/annonex2embl',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.7',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
        ],
    keywords='novel DNA sequences, public sequence databases, European Nucleotide Archive, file conversion, flatfile',
    license='BSD',
    entry_points={
        'console_scripts': [
            'annonex2embl = annonex2embl.CLIOps:start_annonex2embl'
        ],
    },
    packages=['annonex2embl'], # So that the subfolder 'annonex2embl' is read immediately.
    #packages = find_packages(),
    install_requires=['biopython', 'argparse', 'requests', 'unidecode'],
    scripts=glob.glob('scripts/*'),
    test_suite='setup.my_test_suite',
    include_package_data=True,
    zip_safe=False
)
